@@ -1,6 +1,3 @@
-# IMPORTANT: This algorithm is very sensitive and sometimes very slow. Don't make
-#  changes you aren't thoroughly sure about the correctness and performance of.
-
 import itertools
 import math
 from collections.abc import Iterable
@@ -9,7 +6,7 @@ from functools import cache, cached_property
 from types import EllipsisType
 from typing import TYPE_CHECKING, Any, Final, NamedTuple, cast
 
-from ..formatting import FORMATTING_KEYS, get_formatting, is_affected_by_inheriting
+from ..formatting import get_formatting, is_affected_by_inheriting
 from ..helpers import json_str
 from ..prevent_inheritance import prevent_inheritance
 from ..types import (
@@ -70,30 +67,13 @@ def get_cost(formatting_items: Iterable[FormattingItem]):
     return sum(item.cost for item in formatting_items)
 
 
-def set_component_formatting(
-    component: FlatTextComponent, formatting: TextComponentFormatting
-):
-    """Returns a `TextComponentDict` with the content of `component` and the formatting of
-    `formatting`.
-    """
-
-    if isinstance(component, dict):
-        content = {
-            key: value for key, value in component.items() if key not in FORMATTING_KEYS
-        }
-    else:
-        content = {"text": component}
-
-    return cast(TextComponentDict, content | formatting)
-
-
 def factor_common_formatting(subcomponents: list[FlatTextComponent]):
     """Wraps certain ranges of subcomponents into arrays, utilizing array inheritance to
     reduce redundant formatting in the wrapped subcomponents.
 
     ⚠️ Only for use in `minify`. May mutate the inputted subcomponents.
 
-    >>> minify(
+    >>> factor_common_formatting(
     >>>     [
     >>>         {"text": "a", "color": "red"},
     >>>         {"text": "b", "color": "green"},
@@ -167,11 +147,11 @@ def factor_common_formatting(subcomponents: list[FlatTextComponent]):
         best_sublist_factoring: FactoredFormattings | None = None
         best_remainder_factoring: FactoredFormattings | None = None
 
-        # The formattings to consider applying to the sublist.
+        # The formattings to consider ever applying to a sublist.
         potential_formattings: set[FormattingSet] = set()
         # A mapping from each formatting key to the set of formatting items which are in
-        #  `potential_formattings` with that key.
-        potential_formatting_items: dict[str, set[FormattingItem]] = {}
+        #  `potential_formattings` with that key. Excludes keys without any items.
+        potential_items_by_key: dict[str, set[FormattingItem]] = {}
 
         parent_keys = {item.key for item in parent}
 
@@ -186,84 +166,75 @@ def factor_common_formatting(subcomponents: list[FlatTextComponent]):
                 component, {key}
             )
 
-        def get_potential_formattings(
-            # A new component in the sublist that might conflict with the
-            #  `potential_formattings`.
-            potentially_conflicting_component: FlatTextComponent,
-        ):
-            """Gets the updated `potential_formattings`."""
+        def initialize_potential_formattings():
+            first_sublist_formatting = formattings[sublist_start]
+            new_sublist_component = subcomponents[sublist_end - 1]
+            # All formattings but the first that have a chance of being in the sublist.
+            non_first_formattings = formattings[sublist_start + 1 : end]
 
+            # The formatting items to consider factoring.
+            potential_items = {
+                item
+                for item in first_sublist_formatting
+                # Exclude items in the parent, since it's pointless to make a new
+                #  sublist for an item the parent already has.
+                if item not in parent
+                # Exclude items that can only be found once in a sublist.
+                and any(item in formatting for formatting in non_first_formattings)
+                # Exclude items that conflict with the new sublist component.
+                and not formatting_key_affects_component(
+                    item.key, new_sublist_component
+                )
+            }
+
+            for length in range(1, len(potential_items) + 1):
+                combinations = itertools.combinations(potential_items, length)
+                for combination in combinations:
+                    combination_keys: set[str] = set()
+
+                    for item in combination:
+                        combination_keys.add(item.key)
+
+                        if item.key in potential_items_by_key:
+                            potential_items_by_key[item.key].add(item)
+                        else:
+                            potential_items_by_key[item.key] = {item}
+
+                    potential_formatting = {
+                        item
+                        # Every potential formatting should inherit from the parent.
+                        for item in parent
+                        # Exclude items that should be overwritten by the `update`.
+                        if item.key not in combination_keys
+                    }
+                    potential_formatting.update(combination)
+                    potential_formatting = FormattingSet(potential_formatting)
+
+                    potential_formattings.add(potential_formatting)
+
+        def update_potential_formattings():
             if not potential_formattings:
-                # Initialize `potential_formattings`.
+                initialize_potential_formattings()
+                return
 
-                first_sublist_formatting = formattings[sublist_start]
-                # All formattings but the first that have a chance of being in the
-                #  sublist.
-                non_first_formattings = formattings[sublist_start + 1 : end]
-
-                # The formatting items to consider factoring.
-                potential_items = {
-                    item
-                    for item in first_sublist_formatting
-                    # Exclude items in the parent, since it's pointless to make a new
-                    #  sublist for an item the parent already has.
-                    if item not in parent
-                    # Exclude items that can only be found once in the sublist.
-                    and any(item in formatting for formatting in non_first_formattings)
-                    # Exclude items that conflict with the new sublist component.
-                    and not formatting_key_affects_component(
-                        item.key, potentially_conflicting_component
-                    )
-                }
-
-                for length in range(1, len(potential_items) + 1):
-                    combinations = itertools.combinations(potential_items, length)
-                    for combination in combinations:
-                        combination_keys: set[str] = set()
-
-                        for item in combination:
-                            combination_keys.add(item.key)
-
-                            if item.key in potential_formatting_items:
-                                potential_formatting_items[item.key].add(item)
-                            else:
-                                potential_formatting_items[item.key] = {item}
-
-                        potential_formatting = {
-                            item
-                            # Every potential formatting should inherit from the parent.
-                            for item in parent
-                            # Exclude items that will be overwritten.
-                            if item.key not in combination_keys
-                        }
-                        potential_formatting.update(combination)
-                        potential_formattings.add(FormattingSet(potential_formatting))
-
-                return potential_formattings
-
-            # Remove any `potential_formattings` and `potential_formatting_items` that
-            #  conflict with the `potentially_conflicting_component`.
+            # Remove any `potential_formattings` and `potential_items_by_key` that
+            #  conflict with the new sublist component.
 
+            new_sublist_component = subcomponents[sublist_end - 1]
             keys_to_remove = {
                 key
-                for key in potential_formatting_items.keys()
-                if formatting_key_affects_component(
-                    key, potentially_conflicting_component
-                )
+                for key in potential_items_by_key.keys()
+                if formatting_key_affects_component(key, new_sublist_component)
             }
 
             if keys_to_remove:
                 items_to_remove: set[FormattingItem] = set()
                 for key_to_remove in keys_to_remove:
-                    items_to_remove |= potential_formatting_items.pop(key_to_remove)
+                    items_to_remove |= potential_items_by_key.pop(key_to_remove)
 
-                return {
-                    formatting
-                    for formatting in potential_formattings
-                    if not formatting & items_to_remove
-                }
-
-            return potential_formattings
+                for formatting in potential_formattings.copy():
+                    if formatting & items_to_remove:
+                        potential_formattings.remove(formatting)
 
         for sublist_end in range(sublist_start + 1, end + 1):
             sublist_length = sublist_end - sublist_start
@@ -273,15 +244,24 @@ def factor_common_formatting(subcomponents: list[FlatTextComponent]):
                 #  try all the `potential_formattings`.
                 formattings_to_try = {formattings[sublist_start]}
             else:
-                potential_formattings = get_potential_formattings(
-                    potentially_conflicting_component=subcomponents[sublist_end - 1]
-                )
+                update_potential_formattings()
 
                 if not potential_formattings:
                     # There are no more potential formattings that don't conflict with
                     #  any of the sublist's components.
                     break
 
+                last_sublist_formatting = formattings[sublist_end - 1]
+                if not any(
+                    formatting
+                    for formatting in potential_formattings
+                    if formatting <= last_sublist_formatting
+                ):
+                    # The last sublist component wouldn't benefit from these factorings,
+                    #  so an optimal sublist range could just as well exclude this
+                    #  component from the end.
+                    continue
+
                 formattings_to_try = potential_formattings
 
             remainder_factoring = factor_and_get_cost(
